use std::fmt::Display;

use anyhow::{anyhow, Result};
use bytes::Bytes;
use futures::Stream;
use serde::Deserialize;
use tokio_stream::StreamExt;

use crate::health::Health;

#[derive(Clone)]
pub struct Client {
    id: String,
    nats: async_nats::Client,
}

#[derive(Debug)]
pub enum Subject {
    Health,
    Exec,
}

impl Display for Subject {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Subject::Health => write!(f, "health"),
            Subject::Exec => write!(f, "exec"),
        }
    }
}

#[derive(Debug)]
pub struct Message {
    subject: Subject,
    payload: Bytes,
    reply: Option<async_nats::Subject>,
}

impl Message {
    fn from_transport(msg: async_nats::Message) -> Result<Self> {
        let suffix = msg.subject.split_terminator('.').last().unwrap_or_default();

        match suffix {
            "exec" => Ok(Message {
                subject: Subject::Exec,
                payload: msg.payload,
                reply: msg.reply,
            }),
            "health" => Ok(Message {
                subject: Subject::Health,
                payload: msg.payload,
                reply: msg.reply,
            }),
            _ => Err(anyhow!("unknown subject: {}", msg.subject)),
        }
    }

    pub fn subject(&self) -> &Subject {
        &self.subject
    }

    pub fn parse_payload<'a, T: Deserialize<'a>>(&'a self) -> Result<T> {
        Ok(serde_json::from_slice(&self.payload[..])?)
    }

    pub fn health(health: Health) -> Result<Message> {
        Ok(Message {
            subject: Subject::Health,
            payload: Bytes::from_iter(serde_json::to_vec(&health)?),
            reply: None,
        })
    }
}

impl Client {
    pub async fn connect(id: &str) -> Result<Self> {
        let nats = async_nats::ConnectOptions::new()
            .name(format!("Prymn Agent {id}"))
            .custom_inbox_prefix(format!("_INBOX_{id}"))
            .user_and_password("demo_agent".to_owned(), "demo_agent_password".to_owned())
            .connect("localhost")
            .await?;

        tracing::debug!("connected to NATS");

        Ok(Self {
            id: id.to_owned(),
            nats,
        })
    }

    pub async fn publish(&self, msg: Message) {
        let subject = format!("agents.v1.{}.{}", self.id, msg.subject);
        self.nats.publish(subject, msg.payload).await.unwrap();
    }

    pub async fn reply(&self, msg: Message, mut stream: impl Stream<Item = Bytes> + Unpin) {
        match msg.reply {
            Some(reply) => {
                while let Some(data) = stream.next().await {
                    self.nats.publish(reply.clone(), data.into()).await.unwrap();
                }
            }
            None => tracing::warn!(?msg, "tried to reply to message without a reply subject"),
        }
    }

    pub async fn subscribe(&self) -> Result<impl Stream<Item = Message>> {
        let subject = format!("agents.v1.{}.*", self.id);

        let subscriber = self
            .nats
            .subscribe(subject.clone())
            .await?
            .filter_map(move |data| match Message::from_transport(data) {
                Ok(msg) => Some(msg),
                Err(err) => {
                    tracing::warn!("{}", err);
                    None
                }
            });

        tracing::debug!("subscribed on {}", &subject);

        Ok(subscriber)
    }
}
