use std::process::Stdio;

use bytes::Bytes;
use futures::{FutureExt, Stream};
use serde::Deserialize;
use tokio::process::Command;
use tokio_stream::StreamExt;
use tokio_util::codec::{BytesCodec, FramedRead};

use crate::{
    health::Health,
    messaging::{Client, Message, Subject},
};

pub async fn init_services(client: Client) {
    let mut message_stream = client.subscribe().await.unwrap();

    tokio::spawn(async move {
        while let Some(message) = message_stream.next().await {
            let client = client.clone();
            tokio::spawn(async move {
                if let Err(err) = handle_message(client, message).await {
                    tracing::warn!("{err}");
                }
            });
        }
    });
}

async fn handle_message(client: Client, message: Message) -> anyhow::Result<()> {
    match message.subject() {
        Subject::Exec => {
            let stream = exec_handler(message.parse_payload()?).await?;
            client.reply(message, stream).await;
        }
        Subject::Health => {
            let health: Health = message.parse_payload()?;
            tracing::info!(?health, "received a health");
        }
    }

    Ok(())
}

/// An operating system program execution.
#[derive(Debug, Deserialize)]
struct ExecMessage {
    user: String,
    program: String,
    #[serde(default)]
    args: Vec<String>,
}

async fn exec_handler(req: ExecMessage) -> anyhow::Result<impl Stream<Item = Bytes> + Unpin> {
    // TODO: Tasks should be idempontent
    // TODO: Root user should be able to run only specific programs
    let mut cmd = if req.user != "root" {
        let mut cmd = Command::new("sudo");
        cmd.arg("-iu").arg(&req.user).arg("--").arg(&req.program);
        cmd
    } else {
        Command::new(&req.program)
    };

    let mut io = cmd
        .args(&req.args)
        .stdout(Stdio::piped())
        .stderr(Stdio::piped())
        .spawn()?;

    let stdout = FramedRead::new(io.stdout.take().unwrap(), BytesCodec::new())
        .map(|stdout| Bytes::from(stdout.unwrap()));

    let stderr = FramedRead::new(io.stderr.take().unwrap(), BytesCodec::new())
        .map(|stderr| Bytes::from(stderr.unwrap()));

    let exit = async move {
        io.wait()
            .await
            .map(|exit| {
                let exit = exit.to_string();
                Bytes::from(exit)
            })
            .unwrap()
    }
    .into_stream();

    Ok(Box::pin(stdout.merge(stderr).chain(exit)))
}
